use crate::addressing::Address;
use crate::entry::EntryValue;
use crate::error::UpEndError;
use chrono::NaiveDateTime;
use nonempty::NonEmpty;
use std::borrow::Borrow;
use std::convert::TryFrom;
use std::str::FromStr;

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Attribute(pub String);

impl From<&str> for Attribute {
    fn from(str: &str) -> Self {
        Self(str.to_string())
    }
}

#[derive(Debug, Clone, PartialEq, Default)]
pub enum PatternQueryComponent<T>
where
    T: TryFrom<lexpr::Value>,
{
    Exact(T),
    In(Vec<T>),
    Contains(String),
    Variable(Option<String>),
    #[default]
    Discard,
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Provenance(pub String);

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Timestamp(pub NaiveDateTime);

#[derive(Debug, Clone, PartialEq, Default)]
pub struct PatternQuery {
    pub entity: PatternQueryComponent<Address>,
    pub attribute: PatternQueryComponent<Attribute>,
    pub value: PatternQueryComponent<EntryValue>,
    pub provenance: Option<PatternQueryComponent<Provenance>>,
    pub timestamp: Option<PatternQueryComponent<Timestamp>>,
}

impl TryFrom<lexpr::Value> for Address {
    type Error = UpEndError;

    fn try_from(value: lexpr::Value) -> Result<Self, Self::Error> {
        match value {
            lexpr::Value::Symbol(str) => {
                if let Some(address_str) = str.strip_prefix('@') {
                    address_str
                        .parse()
                        .map_err(|e: UpEndError| UpEndError::QueryParseError(e.to_string()))
                } else {
                    Err(UpEndError::QueryParseError(
                        "Incorrect address format (use @address).".into(),
                    ))
                }
            }
            _ => Err(UpEndError::QueryParseError(
                "Incorrect type for address (use @address).".into(),
            )),
        }
    }
}

impl TryFrom<lexpr::Value> for EntryValue {
    type Error = UpEndError;

    fn try_from(value: lexpr::Value) -> Result<Self, Self::Error> {
        match value {
            lexpr::Value::Number(num) => Ok(EntryValue::Number(num.as_f64().ok_or_else(|| {
                UpEndError::QueryParseError(format!("Error processing number ({num:?})."))
            })?)),
            lexpr::Value::Char(chr) => Ok(EntryValue::String(chr.to_string())),
            lexpr::Value::String(str) => Ok(EntryValue::String(str.to_string())),
            lexpr::Value::Symbol(_) => Ok(EntryValue::Address(Address::try_from(value.clone())?)),
            _ => Err(UpEndError::QueryParseError(
                "Value can only be a string, number or address.".into(),
            )),
        }
    }
}

impl TryFrom<lexpr::Value> for Attribute {
    type Error = UpEndError;

    fn try_from(value: lexpr::Value) -> Result<Self, Self::Error> {
        match value {
            lexpr::Value::String(str) => Ok(Attribute(str.to_string())),
            _ => Err(UpEndError::QueryParseError(
                "Can only convert to attribute from string.".into(),
            )),
        }
    }
}

impl TryFrom<lexpr::Value> for Provenance {
    type Error = UpEndError;

    fn try_from(value: lexpr::Value) -> Result<Self, Self::Error> {
        match value {
            lexpr::Value::String(str) => Ok(Provenance(str.to_string())),
            _ => Err(UpEndError::QueryParseError(
                "Can only convert to provenance from string.".into(),
            )),
        }
    }
}

impl TryFrom<lexpr::Value> for Timestamp {
    type Error = UpEndError;

    fn try_from(value: lexpr::Value) -> Result<Self, Self::Error> {
        match value {
            lexpr::Value::Number(num) => {
                if let Some(num) = num.as_i64() {
                    Ok(Timestamp(NaiveDateTime::from_timestamp_opt(num, 0).ok_or(
                        UpEndError::QueryParseError("Couldn't parse timestamp.".into()),
                    )?))
                } else {
                    Err(UpEndError::QueryParseError(
                        "Couldn't parse number as i64.".into(),
                    ))
                }
            }
            _ => Err(UpEndError::QueryParseError(
                "Can only convert to attribute from string.".into(),
            )),
        }
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum QueryPart {
    Matches(PatternQuery),
    Type(String),
}

#[derive(Debug, Clone, PartialEq)]
pub enum QueryQualifier {
    And,
    Or,
    Not,
    Join,
}

#[derive(Debug, Clone, PartialEq)]
pub struct MultiQuery {
    pub qualifier: QueryQualifier,
    pub queries: NonEmpty<Box<Query>>,
}

#[allow(clippy::large_enum_variant)]
#[derive(Debug, Clone, PartialEq)]
pub enum Query {
    SingleQuery(QueryPart),
    MultiQuery(MultiQuery),
}

impl TryFrom<&lexpr::Value> for Query {
    type Error = UpEndError;

    fn try_from(expression: &lexpr::Value) -> Result<Self, Self::Error> {
        fn parse_component<T: TryFrom<lexpr::Value>>(
            value: &lexpr::Value,
        ) -> Result<PatternQueryComponent<T>, UpEndError>
        where
            UpEndError: From<<T as TryFrom<lexpr::Value>>::Error>,
        {
            match value {
                lexpr::Value::Cons(cons) => {
                    if let lexpr::Value::Symbol(symbol) = cons.car() {
                        match symbol.borrow() {
                            "in" => {
                                let (cons_vec, _) = cons.clone().into_vec();
                                if let Some(split) = cons_vec.split_first() {
                                    let args = split.1;
                                    let values: Result<Vec<T>, _> = args
                                        .iter()
                                        .map(|value| T::try_from(value.clone()))
                                        .collect();

                                    Ok(PatternQueryComponent::In(values?))
                                } else {
                                    Err(UpEndError::QueryParseError(
                                        "Malformed expression: Inner value cannot be empty.".into(),
                                    ))
                                }
                            }
                            "contains" => {
                                let (mut cons_vec, _) = cons.clone().into_vec();
                                match cons_vec.len() {
                                    2 => {
                                        let value = cons_vec.remove(1);
                                        if let lexpr::Value::String(str) = value {
                                            Ok(PatternQueryComponent::Contains(str.into_string()))
                                        } else {
                                            Err(UpEndError::QueryParseError("Malformed expression: 'Contains' argument must be a string.".into()))
                                        }
                                    }
                                    _ => Err(UpEndError::QueryParseError(
                                    "Malformed expression: 'Contains' requires a single argument.".into()
                                )),
                                }
                            }
                            _ => Err(UpEndError::QueryParseError(format!(
                                "Malformed expression: Unknown symbol {}",
                                symbol
                            ))),
                        }
                    } else {
                        Err(UpEndError::QueryParseError(format!(
                            "Malformed expression: Inner value '{:?}' is not a symbol.",
                            value
                        )))
                    }
                }
                lexpr::Value::Symbol(symbol) if symbol.starts_with('?') => {
                    let var_name = symbol.strip_prefix('?').unwrap();
                    Ok(PatternQueryComponent::Variable(if var_name.is_empty() {
                        None
                    } else {
                        Some(var_name.into())
                    }))
                }
                lexpr::Value::Symbol(symbol) if symbol.starts_with('_') => {
                    Ok(PatternQueryComponent::Discard)
                }
                _ => Ok(PatternQueryComponent::Exact(T::try_from(value.clone())?)),
            }
        }

        if let lexpr::Value::Cons(value) = expression {
            if let lexpr::Value::Symbol(symbol) = value.car() {
                match symbol.borrow() {
                    "matches" => {
                        let (cons_vec, _) = value.clone().into_vec();
                        if let [_, entity, attribute, value, rest @ ..] = &cons_vec[..] {
                            let entity = parse_component::<Address>(entity)?;
                            let attribute = parse_component::<Attribute>(attribute)?;
                            let value = parse_component::<EntryValue>(value)?;
                            let (provenance, timestamp) = match rest {
                                [] => (None, None),
                                [provenance] => {
                                    (Some(parse_component::<Provenance>(provenance)?), None)
                                }
                                [provenance, timestamp] => (
                                    Some(parse_component::<Provenance>(provenance)?),
                                    Some(parse_component::<Timestamp>(timestamp)?),
                                ),
                                _ => Err(UpEndError::QueryParseError(
                                    "Malformed expression: Too many arguments for `matches`."
                                        .into(),
                                ))?,
                            };
                            Ok(Query::SingleQuery(QueryPart::Matches(PatternQuery {
                                entity,
                                attribute,
                                value,
                                provenance,
                                timestamp,
                            })))
                        } else {
                            Err(UpEndError::QueryParseError(
                                "Malformed expression: Not enough arguments for `matches`.".into(),
                            ))
                        }
                    }
                    "type" => {
                        let (cons_vec, _) = value.clone().into_vec();
                        if let [_, type_name] = &cons_vec[..] {
                            if let lexpr::Value::String(type_name_str) = type_name {
                                Ok(Query::SingleQuery(QueryPart::Type(
                                    type_name_str.to_string(),
                                )))
                            } else {
                                Err(UpEndError::QueryParseError(
                                    "Malformed expression: Type must be specified as a string."
                                        .into(),
                                ))
                            }
                        } else {
                            Err(UpEndError::QueryParseError(
                                "Malformed expression: Wrong number of arguments to 'type'.".into(),
                            ))
                        }
                    }
                    "and" | "or" | "join" => {
                        let (cons_vec, _) = value.clone().into_vec();
                        let sub_expressions = &cons_vec[1..];
                        let values = sub_expressions
                            .iter()
                            .map(|value| Ok(Box::new(Query::try_from(value)?)))
                            .collect::<Result<Vec<Box<Query>>, UpEndError>>()?;

                        if let Some(queries) = NonEmpty::from_vec(values) {
                            Ok(Query::MultiQuery(MultiQuery {
                                qualifier: match symbol.borrow() {
                                    "and" => QueryQualifier::And,
                                    "join" => QueryQualifier::Join,
                                    _ => QueryQualifier::Or,
                                },
                                queries,
                            }))
                        } else {
                            Err(UpEndError::QueryParseError(
                                "Malformed expression: sub-query list cannot be empty.".into(),
                            ))
                        }
                    }
                    "not" => {
                        let (cons_vec, _) = value.clone().into_vec();
                        let sub_expressions = &cons_vec[1..];
                        let values = sub_expressions
                            .iter()
                            .map(|value| Ok(Box::new(Query::try_from(value)?)))
                            .collect::<Result<Vec<Box<Query>>, UpEndError>>()?;

                        if values.len() == 1 {
                            Ok(Query::MultiQuery(MultiQuery {
                                qualifier: QueryQualifier::Not,
                                queries: NonEmpty::from_vec(values).unwrap(),
                            }))
                        } else {
                            Err(UpEndError::QueryParseError(
                                "Malformed expression: NOT takes exactly one parameter.".into(),
                            ))
                        }
                    }
                    _ => Err(UpEndError::QueryParseError(format!(
                        "Malformed expression: Unknown symbol '{}'.",
                        symbol
                    ))),
                }
            } else {
                Err(UpEndError::QueryParseError(format!(
                    "Malformed expression: Value '{:?}' is not a symbol.",
                    value
                )))
            }
        } else {
            Err(UpEndError::QueryParseError(
                "Malformed expression: Not a list.".into(),
            ))
        }
    }
}

impl FromStr for Query {
    type Err = UpEndError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let sexp = lexpr::from_str_custom(s, lexpr::parse::Options::new()).map_err(|e| {
            UpEndError::QueryParseError(format!("failed to parse s-expression: {e}"))
        })?;
        Query::try_from(&sexp)
    }
}

#[cfg(test)]
mod test {
    use crate::error::UpEndError;

    use super::*;
    use url::Url;

    #[test]
    fn test_matches() -> Result<(), UpEndError> {
        let query = "(matches ? ? ?)".parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::Variable(None),
                ..Default::default()
            }))
        );

        let address = Address::Url(Url::parse("https://upend.dev").unwrap());
        let query = format!("(matches @{address} ? ?)").parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Exact(address),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::Variable(None),
                ..Default::default()
            }))
        );

        let query = r#"(matches ? "FOO" ?)"#.parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Exact("FOO".into()),
                value: PatternQueryComponent::Variable(None),
                ..Default::default()
            }))
        );

        let value = EntryValue::Number(1337.93);
        let query = "(matches ? ? 1337.93)".parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::Exact(value),
                ..Default::default()
            }))
        );

        Ok(())
    }

    #[test]
    fn test_joins() -> Result<(), UpEndError> {
        let query = "(matches ?a ?b ?)".parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(Some("a".into())),
                attribute: PatternQueryComponent::Variable(Some("b".into())),
                value: PatternQueryComponent::Variable(None),
                ..Default::default()
            }))
        );

        Ok(())
    }

    #[test]
    fn test_in_parse() -> Result<(), UpEndError> {
        let query = r#"(matches ? (in "FOO" "BAR") ?)"#.parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::In(vec!("FOO".into(), "BAR".into())),
                value: PatternQueryComponent::Variable(None),
                ..Default::default()
            }))
        );

        let values: Vec<EntryValue> = vec!["FOO".into(), "BAR".into()];
        let query = r#"(matches ? ? (in "FOO" "BAR"))"#.parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::In(values),
                ..Default::default()
            }))
        );

        let values = vec![EntryValue::Number(1337.93), EntryValue::Number(1968.12)];
        let query = r#"(matches ? ? (in 1337.93 1968.12))"#.parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::In(values),
                ..Default::default()
            }))
        );

        // Invalid queries
        let values: Vec<EntryValue> = vec!["FOO".into(), EntryValue::Number(1337.93)];
        let query = r#"(matches ? ? (in "FOO" 1337.93))"#.parse::<Query>()?;

        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::In(values),
                ..Default::default()
            }))
        );

        let values = vec![EntryValue::Number(1337.93), "FOO".into()];
        let query = r#"(matches ? ? (in 1337.93 "FOO"))"#.parse::<Query>()?;

        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::In(values),
                ..Default::default()
            }))
        );

        Ok(())
    }

    #[test]
    fn test_contains() -> Result<(), UpEndError> {
        let query = r#"(matches (contains "foo") ? ?)"#.parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Contains("foo".to_string()),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::Variable(None),
                ..Default::default()
            }))
        );

        let query = r#"(matches ? (contains "foo") ?)"#.parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Contains("foo".to_string()),
                value: PatternQueryComponent::Variable(None),
                ..Default::default()
            }))
        );

        let query = r#"(matches ? ? (contains "foo"))"#.parse::<Query>()?;
        assert_eq!(
            query,
            Query::SingleQuery(QueryPart::Matches(PatternQuery {
                entity: PatternQueryComponent::Variable(None),
                attribute: PatternQueryComponent::Variable(None),
                value: PatternQueryComponent::Contains("foo".to_string()),
                ..Default::default()
            }))
        );

        Ok(())
    }
}
